@@ -28,70 +28,15 @@ import (
 	"github.com/prometheus/common/model"
 
 	"github.com/prometheus/alertmanager/cluster"
-	"github.com/prometheus/alertmanager/config"
+	"github.com/prometheus/alertmanager/inhibit"
 	"github.com/prometheus/alertmanager/nflog"
 	"github.com/prometheus/alertmanager/nflog/nflogpb"
 	"github.com/prometheus/alertmanager/silence"
-	"github.com/prometheus/alertmanager/template"
 	"github.com/prometheus/alertmanager/types"
 )
 
-var (
-	numNotifications = prometheus.NewCounterVec(prometheus.CounterOpts{
-		Namespace: "alertmanager",
-		Name:      "notifications_total",
-		Help:      "The total number of attempted notifications.",
-	}, []string{"integration"})
-
-	numFailedNotifications = prometheus.NewCounterVec(prometheus.CounterOpts{
-		Namespace: "alertmanager",
-		Name:      "notifications_failed_total",
-		Help:      "The total number of failed notifications.",
-	}, []string{"integration"})
-
-	notificationLatencySeconds = prometheus.NewHistogramVec(prometheus.HistogramOpts{
-		Namespace: "alertmanager",
-		Name:      "notification_latency_seconds",
-		Help:      "The latency of notifications in seconds.",
-		Buckets:   []float64{1, 5, 10, 15, 20},
-	}, []string{"integration"})
-)
-
-func init() {
-	numNotifications.WithLabelValues("email")
-	numNotifications.WithLabelValues("hipchat")
-	numNotifications.WithLabelValues("pagerduty")
-	numNotifications.WithLabelValues("wechat")
-	numNotifications.WithLabelValues("pushover")
-	numNotifications.WithLabelValues("slack")
-	numNotifications.WithLabelValues("opsgenie")
-	numNotifications.WithLabelValues("webhook")
-	numNotifications.WithLabelValues("victorops")
-	numFailedNotifications.WithLabelValues("email")
-	numFailedNotifications.WithLabelValues("hipchat")
-	numFailedNotifications.WithLabelValues("pagerduty")
-	numFailedNotifications.WithLabelValues("wechat")
-	numFailedNotifications.WithLabelValues("pushover")
-	numFailedNotifications.WithLabelValues("slack")
-	numFailedNotifications.WithLabelValues("opsgenie")
-	numFailedNotifications.WithLabelValues("webhook")
-	numFailedNotifications.WithLabelValues("victorops")
-	notificationLatencySeconds.WithLabelValues("email")
-	notificationLatencySeconds.WithLabelValues("hipchat")
-	notificationLatencySeconds.WithLabelValues("pagerduty")
-	notificationLatencySeconds.WithLabelValues("wechat")
-	notificationLatencySeconds.WithLabelValues("pushover")
-	notificationLatencySeconds.WithLabelValues("slack")
-	notificationLatencySeconds.WithLabelValues("opsgenie")
-	notificationLatencySeconds.WithLabelValues("webhook")
-	notificationLatencySeconds.WithLabelValues("victorops")
-
-	prometheus.MustRegister(numNotifications)
-	prometheus.MustRegister(numFailedNotifications)
-	prometheus.MustRegister(notificationLatencySeconds)
-}
-
-type notifierConfig interface {
+// ResolvedSender returns true if resolved notifications should be sent.
+type ResolvedSender interface {
 	SendResolved() bool
 }
 
@@ -99,6 +44,52 @@ type notifierConfig interface {
 // to a notification pipeline.
 const MinTimeout = 10 * time.Second
 
+// Notifier notifies about alerts under constraints of the given context. It
+// returns an error if unsuccessful and a flag whether the error is
+// recoverable. This information is useful for a retry logic.
+type Notifier interface {
+	Notify(context.Context, ...*types.Alert) (bool, error)
+}
+
+// Integration wraps a notifier and its configuration to be uniquely identified
+// by name and index from its origin in the configuration.
+type Integration struct {
+	notifier Notifier
+	rs       ResolvedSender
+	name     string
+	idx      int
+}
+
+// NewIntegration returns a new integration.
+func NewIntegration(notifier Notifier, rs ResolvedSender, name string, idx int) Integration {
+	return Integration{
+		notifier: notifier,
+		rs:       rs,
+		name:     name,
+		idx:      idx,
+	}
+}
+
+// Notify implements the Notifier interface.
+func (i *Integration) Notify(ctx context.Context, alerts ...*types.Alert) (bool, error) {
+	return i.notifier.Notify(ctx, alerts...)
+}
+
+// SendResolved implements the ResolvedSender interface.
+func (i *Integration) SendResolved() bool {
+	return i.rs.SendResolved()
+}
+
+// Name returns the name of the integration.
+func (i *Integration) Name() string {
+	return i.name
+}
+
+// Index returns the index of the integration.
+func (i *Integration) Index() int {
+	return i.idx
+}
+
 // notifyKey defines a custom type with which a context is populated to
 // avoid accidental collisions.
 type notifyKey int
@@ -162,14 +153,6 @@ func ReceiverName(ctx context.Context) (string, bool) {
 	return v, ok
 }
 
-func receiverName(ctx context.Context, l log.Logger) string {
-	recv, ok := ReceiverName(ctx)
-	if !ok {
-		level.Error(l).Log("msg", "Missing receiver")
-	}
-	return recv
-}
-
 // GroupKey extracts a group key from the context. Iff none exists, the
 // second argument is false.
 func GroupKey(ctx context.Context) (string, bool) {
@@ -177,14 +160,6 @@ func GroupKey(ctx context.Context) (string, bool) {
 	return v, ok
 }
 
-func groupLabels(ctx context.Context, l log.Logger) model.LabelSet {
-	groupLabels, ok := GroupLabels(ctx)
-	if !ok {
-		level.Error(l).Log("msg", "Missing group labels")
-	}
-	return groupLabels
-}
-
 // GroupLabels extracts grouping label set from the context. Iff none exists, the
 // second argument is false.
 func GroupLabels(ctx context.Context) (model.LabelSet, bool) {
@@ -231,43 +206,101 @@ type NotificationLog interface {
 	Query(params ...nflog.QueryParam) ([]*nflogpb.Entry, error)
 }
 
-// BuildPipeline builds a map of receivers to Stages.
-func BuildPipeline(
-	confs []*config.Receiver,
-	tmpl *template.Template,
+type metrics struct {
+	numNotifications           *prometheus.CounterVec
+	numFailedNotifications     *prometheus.CounterVec
+	notificationLatencySeconds *prometheus.HistogramVec
+}
+
+func newMetrics(r prometheus.Registerer) *metrics {
+	m := &metrics{
+		numNotifications: prometheus.NewCounterVec(prometheus.CounterOpts{
+			Namespace: "alertmanager",
+			Name:      "notifications_total",
+			Help:      "The total number of attempted notifications.",
+		}, []string{"integration"}),
+		numFailedNotifications: prometheus.NewCounterVec(prometheus.CounterOpts{
+			Namespace: "alertmanager",
+			Name:      "notifications_failed_total",
+			Help:      "The total number of failed notifications.",
+		}, []string{"integration"}),
+		notificationLatencySeconds: prometheus.NewHistogramVec(prometheus.HistogramOpts{
+			Namespace: "alertmanager",
+			Name:      "notification_latency_seconds",
+			Help:      "The latency of notifications in seconds.",
+			Buckets:   []float64{1, 5, 10, 15, 20},
+		}, []string{"integration"}),
+	}
+	for _, integration := range []string{
+		"email",
+		"hipchat",
+		"pagerduty",
+		"wechat",
+		"pushover",
+		"slack",
+		"opsgenie",
+		"webhook",
+		"victorops",
+	} {
+		m.numNotifications.WithLabelValues(integration)
+		m.numFailedNotifications.WithLabelValues(integration)
+		m.notificationLatencySeconds.WithLabelValues(integration)
+	}
+	r.MustRegister(m.numNotifications, m.numFailedNotifications, m.notificationLatencySeconds)
+	return m
+}
+
+type PipelineBuilder struct {
+	metrics *metrics
+}
+
+func NewPipelineBuilder(r prometheus.Registerer) *PipelineBuilder {
+	return &PipelineBuilder{
+		metrics: newMetrics(r),
+	}
+}
+
+// New returns a map of receivers to Stages.
+func (pb *PipelineBuilder) New(
+	receivers map[string][]Integration,
 	wait func() time.Duration,
-	muter types.Muter,
-	silences *silence.Silences,
+	inhibitor *inhibit.Inhibitor,
+	silencer *silence.Silencer,
 	notificationLog NotificationLog,
-	marker types.Marker,
 	peer *cluster.Peer,
-	logger log.Logger,
 ) RoutingStage {
-	rs := RoutingStage{}
+	rs := make(RoutingStage, len(receivers))
 
 	ms := NewGossipSettleStage(peer)
-	is := NewInhibitStage(muter)
-	ss := NewSilenceStage(silences, marker)
+	is := NewMuteStage(inhibitor)
+	ss := NewMuteStage(silencer)
 
-	for _, rc := range confs {
-		rs[rc.Name] = MultiStage{ms, is, ss, createStage(rc, tmpl, wait, notificationLog, logger)}
+	for name := range receivers {
+		st := createReceiverStage(name, receivers[name], wait, notificationLog, pb.metrics)
+		rs[name] = MultiStage{ms, is, ss, st}
 	}
 	return rs
 }
 
-// createStage creates a pipeline of stages for a receiver.
-func createStage(rc *config.Receiver, tmpl *template.Template, wait func() time.Duration, notificationLog NotificationLog, logger log.Logger) Stage {
+// createReceiverStage creates a pipeline of stages for a receiver.
+func createReceiverStage(
+	name string,
+	integrations []Integration,
+	wait func() time.Duration,
+	notificationLog NotificationLog,
+	metrics *metrics,
+) Stage {
 	var fs FanoutStage
-	for _, i := range BuildReceiverIntegrations(rc, tmpl, logger) {
+	for i := range integrations {
 		recv := &nflogpb.Receiver{
-			GroupName:   rc.Name,
-			Integration: i.name,
-			Idx:         uint32(i.idx),
+			GroupName:   name,
+			Integration: integrations[i].Name(),
+			Idx:         uint32(integrations[i].Index()),
 		}
 		var s MultiStage
 		s = append(s, NewWaitStage(wait))
-		s = append(s, NewDedupStage(i, notificationLog, recv))
-		s = append(s, NewRetryStage(i, rc.Name))
+		s = append(s, NewDedupStage(&integrations[i], notificationLog, recv))
+		s = append(s, NewRetryStage(integrations[i], name, metrics))
 		s = append(s, NewSetNotifiesStage(notificationLog, recv))
 
 		fs = append(fs, s)
@@ -294,7 +327,7 @@ func (rs RoutingStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.
 	return s.Exec(ctx, l, alerts...)
 }
 
-// A MultiStage executes a series of stages sequencially.
+// A MultiStage executes a series of stages sequentially.
 type MultiStage []Stage
 
 // Exec implements the Stage interface.
@@ -329,7 +362,14 @@ func (fs FanoutStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.A
 		go func(s Stage) {
 			if _, _, err := s.Exec(ctx, l, alerts...); err != nil {
 				me.Add(err)
-				level.Error(l).Log("msg", "Error on notify", "err", err)
+				lvl := level.Error(l)
+				if ctx.Err() == context.Canceled {
+					// It is expected for the context to be canceled on
+					// configuration reload or shutdown. In this case, the
+					// message should only be logged at the debug level.
+					lvl = level.Debug(l)
+				}
+				lvl.Log("msg", "Error on notify", "err", err, "context_err", ctx.Err())
 			}
 			wg.Done()
 		}(s)
@@ -359,72 +399,27 @@ func (n *GossipSettleStage) Exec(ctx context.Context, l log.Logger, alerts ...*t
 	return ctx, alerts, nil
 }
 
-// InhibitStage filters alerts through an inhibition muter.
-type InhibitStage struct {
+// MuteStage filters alerts through a Muter.
+type MuteStage struct {
 	muter types.Muter
 }
 
-// NewInhibitStage return a new InhibitStage.
-func NewInhibitStage(m types.Muter) *InhibitStage {
-	return &InhibitStage{muter: m}
+// NewMuteStage return a new MuteStage.
+func NewMuteStage(m types.Muter) *MuteStage {
+	return &MuteStage{muter: m}
 }
 
 // Exec implements the Stage interface.
-func (n *InhibitStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
+func (n *MuteStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
 	var filtered []*types.Alert
 	for _, a := range alerts {
 		// TODO(fabxc): increment total alerts counter.
-		// Do not send the alert if the silencer mutes it.
+		// Do not send the alert if muted.
 		if !n.muter.Mutes(a.Labels) {
-			// TODO(fabxc): increment muted alerts counter.
 			filtered = append(filtered, a)
 		}
+		// TODO(fabxc): increment muted alerts counter if muted.
 	}
-
-	return ctx, filtered, nil
-}
-
-// SilenceStage filters alerts through a silence muter.
-type SilenceStage struct {
-	silences *silence.Silences
-	marker   types.Marker
-}
-
-// NewSilenceStage returns a new SilenceStage.
-func NewSilenceStage(s *silence.Silences, mk types.Marker) *SilenceStage {
-	return &SilenceStage{
-		silences: s,
-		marker:   mk,
-	}
-}
-
-// Exec implements the Stage interface.
-func (n *SilenceStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Alert) (context.Context, []*types.Alert, error) {
-	var filtered []*types.Alert
-	for _, a := range alerts {
-		// TODO(fabxc): increment total alerts counter.
-		// Do not send the alert if the silencer mutes it.
-		sils, err := n.silences.Query(
-			silence.QState(types.SilenceStateActive),
-			silence.QMatches(a.Labels),
-		)
-		if err != nil {
-			level.Error(l).Log("msg", "Querying silences failed", "err", err)
-		}
-
-		if len(sils) == 0 {
-			// TODO(fabxc): increment muted alerts counter.
-			filtered = append(filtered, a)
-			n.marker.SetSilenced(a.Labels.Fingerprint())
-		} else {
-			ids := make([]string, len(sils))
-			for i, s := range sils {
-				ids[i] = s.Id
-			}
-			n.marker.SetSilenced(a.Labels.Fingerprint(), ids...)
-		}
-	}
-
 	return ctx, filtered, nil
 }
 
@@ -454,20 +449,20 @@ func (ws *WaitStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Al
 // DedupStage filters alerts.
 // Filtering happens based on a notification log.
 type DedupStage struct {
+	rs    ResolvedSender
 	nflog NotificationLog
 	recv  *nflogpb.Receiver
-	conf  notifierConfig
 
 	now  func() time.Time
 	hash func(*types.Alert) uint64
 }
 
 // NewDedupStage wraps a DedupStage that runs against the given notification log.
-func NewDedupStage(i Integration, l NotificationLog, recv *nflogpb.Receiver) *DedupStage {
+func NewDedupStage(rs ResolvedSender, l NotificationLog, recv *nflogpb.Receiver) *DedupStage {
 	return &DedupStage{
+		rs:    rs,
 		nflog: l,
 		recv:  recv,
-		conf:  i.conf,
 		now:   utcNow,
 		hash:  hashAlert,
 	}
@@ -540,7 +535,7 @@ func (n *DedupStage) needsUpdate(entry *nflogpb.Entry, firing, resolved map[uint
 		return len(entry.FiringAlerts) > 0
 	}
 
-	if n.conf.SendResolved() && !entry.IsResolvedSubset(resolved) {
+	if n.rs.SendResolved() && !entry.IsResolvedSubset(resolved) {
 		return true
 	}
 
@@ -604,13 +599,15 @@ func (n *DedupStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Al
 type RetryStage struct {
 	integration Integration
 	groupName   string
+	metrics     *metrics
 }
 
 // NewRetryStage returns a new instance of a RetryStage.
-func NewRetryStage(i Integration, groupName string) *RetryStage {
+func NewRetryStage(i Integration, groupName string, metrics *metrics) *RetryStage {
 	return &RetryStage{
 		integration: i,
 		groupName:   groupName,
+		metrics:     metrics,
 	}
 }
 
@@ -621,7 +618,7 @@ func (r RetryStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Ale
 	// If we shouldn't send notifications for resolved alerts, but there are only
 	// resolved alerts, report them all as successfully notified (we still want the
 	// notification log to log them for the next run of DedupStage).
-	if !r.integration.conf.SendResolved() {
+	if !r.integration.SendResolved() {
 		firing, ok := FiringAlerts(ctx)
 		if !ok {
 			return ctx, nil, fmt.Errorf("firing alerts missing")
@@ -663,13 +660,13 @@ func (r RetryStage) Exec(ctx context.Context, l log.Logger, alerts ...*types.Ale
 		case <-tick.C:
 			now := time.Now()
 			retry, err := r.integration.Notify(ctx, sent...)
-			notificationLatencySeconds.WithLabelValues(r.integration.name).Observe(time.Since(now).Seconds())
-			numNotifications.WithLabelValues(r.integration.name).Inc()
+			r.metrics.notificationLatencySeconds.WithLabelValues(r.integration.Name()).Observe(time.Since(now).Seconds())
+			r.metrics.numNotifications.WithLabelValues(r.integration.Name()).Inc()
 			if err != nil {
-				numFailedNotifications.WithLabelValues(r.integration.name).Inc()
-				level.Debug(l).Log("msg", "Notify attempt failed", "attempt", i, "integration", r.integration.name, "receiver", r.groupName, "err", err)
+				r.metrics.numFailedNotifications.WithLabelValues(r.integration.Name()).Inc()
+				level.Debug(l).Log("msg", "Notify attempt failed", "attempt", i, "integration", r.integration.Name(), "receiver", r.groupName, "err", err)
 				if !retry {
-					return ctx, alerts, fmt.Errorf("cancelling notify retry for %q due to unrecoverable error: %s", r.integration.name, err)
+					return ctx, alerts, fmt.Errorf("cancelling notify retry for %q due to unrecoverable error: %s", r.integration.Name(), err)
 				}
 
 				// Save this error to be able to return the last seen error by an
